#include <X11/Xlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

static const char *dispname=":0";

extern Display *XOpenDisplay(const char*);

unsigned char   bbuf[8192];
unsigned short  wbuf[8192];
unsigned short  trans[256];

#define SIZEX  (9*16*2)
#define SIZEY  (284)
#define FORMAT XYBitmap
#define DEPTH  1
#define PAD    8

void preparetrans(void) {
  int i;
  for(i=0; i<256; i++) {
    trans[i] = 0xffff;
    if(i & 0x01) trans[i] ^= (3 << 0x00);
    if(i & 0x02) trans[i] ^= (3 << 0x02);
    if(i & 0x04) trans[i] ^= (3 << 0x04);
    if(i & 0x08) trans[i] ^= (3 << 0x06);
    if(i & 0x10) trans[i] ^= (3 << 0x08);
    if(i & 0x20) trans[i] ^= (3 << 0x0A);
    if(i & 0x40) trans[i] ^= (3 << 0x0C);
    if(i & 0x80) trans[i] ^= (3 << 0x0E);
  }
}

void bufupdate(void) {
}

int main(int ac,char *av[])
{
  Display   *disp;
  Window     win;
  XEvent     evt;
  XGCValues  gcval;
  GC         gc;
  XImage    *ximg;
  long       emask;
  int        i, vfd, count, stop=0;
  char      *arvdev=ARVDEV;

  if(ac > 2) {
    printf("\nusage: xarvideo [file]\n");
    return 1;
  }

  if(ac == 2) arvdev=av[1];

  vfd = open(arvdev, O_RDONLY);
  if(vfd < 0) {
    printf("\nERROR: can't open file %s\n", arvdev);
    return 2;
  }

  preparetrans();

  disp = XOpenDisplay(dispname);
  win  = XCreateSimpleWindow(disp,
			     DefaultRootWindow(disp),
			     100,100,
			     SIZEX,SIZEY,
			     1,
			     BlackPixel(disp,DefaultScreen(disp)),
			     WhitePixel(disp,DefaultScreen(disp)));

  emask = ResizeRedirectMask | ExposureMask | ButtonPressMask | ButtonReleaseMask |
          PointerMotionMask | KeyPressMask | KeyReleaseMask | FocusChangeMask;
  
  XSelectInput(disp,win,emask);
  XMapWindow(disp,win);
  XStoreName(disp,win,"ARVID video input");

  /* set necessary GC values */
  gcval.foreground = BlackPixel(disp,DefaultScreen(disp));
  gcval.background = WhitePixel(disp,DefaultScreen(disp));
  gc = XCreateGC(disp,win,GCForeground | GCBackground,&gcval);

  XInitImage(ximg);
  ximg = XCreateImage(disp,
		      CopyFromParent,
		      DEPTH,
		      FORMAT,
		      0,
		      (char *) wbuf,
		      SIZEX,
		      SIZEY,
		      PAD,
		      SIZEX/8);

  while(1) {

    if( XCheckWindowEvent(disp,win,emask,&evt) == True ) {
      switch(evt.type) {
      case FocusIn:
	break;
      case FocusOut:
	break;
      case MotionNotify:
	break;
      case ButtonPress:
/*      printf("Button = %04x\n",evt.xbutton.button);
	printf("State  = %04x\n",evt.xbutton.state);  */
	if(evt.xbutton.button == 1) stop = (stop)?0:1;
	if(evt.xbutton.button == 2) return 0;
	if(evt.xbutton.button == 3) {
	  lseek(vfd, 0, SEEK_SET);
	  stop=0;
	}
	break;
      case ButtonRelease:
	break;
      case KeyPress:
/*      printf("Key   = %04x\n",evt.xkey.keycode);
	printf("State = %04x\n",evt.xkey.state);   */
	break;
      case KeyRelease:
	break;
      case ResizeRequest:
	XResizeWindow(disp,win,SIZEX,SIZEY);
	break;
      case Expose:
	XPutImage(disp,win,gc,ximg,0,0,0,0,SIZEX,SIZEY);
	break;
      }
    }
    if(stop) continue;
    count=read(vfd, bbuf, 8192);
    for(i=0; i<8192; i++) wbuf[i] = trans[bbuf[i]];
    if(count>0) XPutImage(disp,win,gc,ximg,0,0,0,0,SIZEX,SIZEY);
    usleep(5);
  }

  XCloseDisplay(disp);
}






