/*
 *  $ReleaseID$
 *
 *  $Id$
 */

#include <unistd.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/io.h>
#include "arvid.h"
#include "arv-data.h"
#include "arv-ll.h"
#include "arv-io.h"
#include "arv-isr.h"
#include "tracing.h"

volatile word currmode=0;
static char stat_text[16384];

void arv_rque() {
  port_cmd(currmode | RQUE);
  port_cmd(currmode);
  hwbptr=bufptr=0;
  bufnum=0;
}

void arv_initstat() {
  arvid.fr_lost_total = 0;
  arvid.fr_repeat_total = 0;
}

static void arv_setup(void) {
  arvid.buffers = 8;
  arvid.bufflen = 4096*2;
  arvid.ircblen = 312*2*12;
  arvid.vidblen = 284*2*9;

  arv_initstat();
  arvid.phase = 4;

  switch(arvid.model) {
  case UNDETECTED:
  case ARV1030:
    arvid.modidx = 1030;
    break;
  case ARV1031:
    arvid.modidx = 1031;
    break;
  case ARV1051:
    arvid.modidx = 1051;
    arvid.buffers = 16;
    arvid.vidblen = 296*2*13;
    break;
  case ARV1052:
    arvid.modidx = 1052;
    arvid.buffers = 64;
    arvid.vidblen = 296*2*13;
  }
}

static void arv_detect(void) {
  currmode=(DATA | SEND | TOPC);
  arv_rque();

  arvid.model=UNDETECTED;
  port_cmd(currmode | (1<<9));
  if(port_stat() & 0x10) arvid.model=ARV1052;
  port_cmd(currmode | (1<<8));
  if(port_stat() & 0x10) arvid.model=ARV1051;
  port_cmd(currmode | (1<<2));
  if(port_stat() & 0x10) arvid.model=ARV1031;
  else arvid.model=ARV1030;
  arv_setup();
}

int _arvid_init(void) {
#ifdef MODULE
  arvid.iobase  = arvparam[0];
  arvid.irq     = arvparam[1];
#else
  arvid.iobase  = ARVID_IOBASE;
  arvid.irq     = ARVID_IRQ;
#endif
  arv_detect();
  if(arv_ramtest()) return ARVERR_BADRAM;
  return ARVERR_OK;
}

void _arvid_reset() {
  currmode=(DATA | SEND | TOPC);
  arv_rque();
}

int _arvid_get(void) {
/*  int err=request_irq(arvid.irq, irq_hdlr, SA_INTERRUPT, "arvid", arvid_id); */
  int i,err=0;
  arvid.irq_lost = 0;
  if(request_irq(arvid.irq, irq_hdlr, 0, "arvid", arvid_id)) err=-1;
  if(err) return err;
  for(i=0; i<DRV_BUFNUM; i++)
    if((buffer[i]=kmalloc(arvid.bufflen,GFP_KERNEL)) == NULL) err = -1;
  return err;
}

void _arvid_free(void) {
  int i;
  free_irq(arvid.irq, arvid_id);
  for(i=0; i<DRV_BUFNUM; i++) kfree(buffer[i]);
}

int _arvid_open() {
  real_hdlr=NIL;
  arv_rque();
  arv_ramclear();
  bufnum=0;
  return 0;
}

int _arvid_close() {
  while(bufnum != 0) {}
  arv_rque();
  bufnum=0;
  return 0;
}

int _arvid_read(char *buff, int req_len) {
  int i,bptr,bnum,result=0;
  TRACE_FUN(5, "_arvid_read");

  if((currmode & RECV) == 0) {
    while(bufnum != 0) {}
    currmode |= RECV;
    arv_rque();
  }
  result = verify_area(VERIFY_WRITE, buff, req_len);
  if (result) {
    TRACEx1(1, "verify_area failed, exitcode = %d", result);
    TRACE_EXIT;
    return -EIO;
  }

  bptr = bufptr;
  bnum = req_len/arvid.bufflen;
  bufnum = bnum;

  real_hdlr=read_hdlr;

  /* while(bufnum != 0) { interruptible_sleep_on(&arvid_waitq); } */

  i=0;
  while(bufnum != 0) {
    interruptible_sleep_on(&arvid_waitq);
    for(; i<bnum-bufnum; i++) {
      /* drop 1 first word in each buffer because h/w add it to begin */
      memcpy_tofs(buff+i*arvid.bufflen, buffer[bptr], arvid.bufflen);
      bptr = ++bptr % DRV_BUFNUM;
    }
  }

  TRACE_EXIT;
  return bnum*arvid.bufflen;
}

int _arvid_write(const char *buff, int req_len) {
  int i,bptr,bnum,result=0;
  TRACE_FUN(5, "_arvid_write");

  if((currmode & RECV) != 0) {
    while(bufnum != 0) {}
    currmode ^= RECV;
    arv_rque();
  }
  result = verify_area(VERIFY_READ, buff, req_len);
  if (result) {
    TRACEx1(1, "verify_area failed, exitcode = %d", result);
    TRACE_EXIT;
    return -EIO;
  }

  /* <+++ optimisation area */
  bnum = req_len/arvid.bufflen;
  while(bufnum+bnum > DRV_BUFNUM) { interruptible_sleep_on(&arvid_waitq); }

  real_hdlr=NIL;
  bptr = (bufptr+bufnum) % DRV_BUFNUM;
  bufnum += bnum;
  for(i=0; i<bnum; i++) {
    memcpy_fromfs(buffer[bptr], buff+i*arvid.bufflen, arvid.bufflen);
    bptr = ++bptr % DRV_BUFNUM;
  }

  real_hdlr=write_hdlr;
  /* <--- optimisation area */

  TRACE_EXIT;
  return bnum*arvid.bufflen;
}

int _arvid_write_new(const char *buff, int req_len) {
  int i,bptr,bnum,result=0;
  TRACE_FUN(5, "_arvid_write");

  if((currmode & RECV) != 0) {
    while(bufnum != 0) {}
    currmode ^= RECV;
    arv_rque();
  }
  result = verify_area(VERIFY_READ, buff, req_len);
  if (result) {
    TRACEx1(1, "verify_area failed, exitcode = %d", result);
    TRACE_EXIT;
    return -EIO;
  }

  /* <+++ optimisation area */
  bnum = req_len/arvid.bufflen;

  real_hdlr=NIL;
  bptr = (bufptr+bufnum) % DRV_BUFNUM;

  for(i=0; i<bnum; i++) {
    memcpy_fromfs(buffer[bptr], buff+i*arvid.bufflen, arvid.bufflen);
    bptr = ++bptr % DRV_BUFNUM;
    bufnum++;
  }

  real_hdlr=write_hdlr;

  while(bufnum+bnum > DRV_BUFNUM) {
    interruptible_sleep_on(&arvid_waitq);
  }

  /* <--- optimisation area */

  TRACE_EXIT;
  return bnum*arvid.bufflen;
}

void arv_statupdate() {
  if(!arvid.busy) { 
    arvid.fr_lost = 0;
    arvid.fr_repeat = 0;
    return;
  }
  if(real_hdlr==NIL) {
    if(currmode & RECV) {
      arvid.fr_lost++;
      arvid.fr_lost_total++;
    } else {
      arvid.fr_repeat++;
      arvid.fr_repeat_total++;
    }
  } 
}

int _arvid_status(char *buff, int req_len) {
  int len,result=0;
  TRACE_FUN(5, "_arvid_status");
  result=verify_area(VERIFY_WRITE, buff, req_len);
  if (result) {
    TRACEx1(1, "verify_area failed, exitcode = %d", result);
    TRACE_EXIT;
    return -EIO;
  }
  len=sprintf(stat_text,
	      "Arvid driver info\n"
	      "-----------------\n"
	      "\n"
	      "model : Arvid-%d\n"
	      "port  : 0x%03x\n"
	      "irq   : %d\n"
	      "\n"
	      "lost irq : %d\n"
	      "\n"
	      "lost frames           : %d\n"
	      "lost frames total     : %d\n"
	      "repeated frames       : %d\n"
	      "repeated frames total : %d\n"
	      "\n"
	      "current phase : %d\n"
	      "\n"
	      , arvid.modidx, arvid.iobase, arvid.irq, arvid.irq_lost,
	      arvid.fr_lost, arvid.fr_lost_total,
	      arvid.fr_repeat, arvid.fr_repeat_total,
	      arvid.phase);

  if(len>req_len) len=req_len;
  memcpy_tofs(buff, stat_text, len);
  TRACE_EXIT;
  return len;
}

/*
  if(waserr) {
    TRACE(2, "interrupted by non-blockable signal");
    TRACE_EXIT;
    return -EIO;
  }
*/

void arv_bufclear(void) {
  int i,j;
  for(i=0; i<DRV_BUFNUM; i++) for(j=0; j<arvid.bufflen; j++) *(buffer[i]+j)=0;
}

void arv_ramclear() {
  int i,bnum;
  static byte *dataout;
  dataout=kmalloc(arvid.bufflen,GFP_KERNEL);
  for(i=0; i<arvid.bufflen; i++) dataout[i]=0;
  for(bnum=0; bnum<arvid.buffers; bnum++) wbuf(bnum,dataout);
  kfree(dataout);
}

int arv_ramtest() {
  int i,bnum;
  int err = 0;
  byte *datainp, *dataout;
  
  datainp=kmalloc(arvid.bufflen,GFP_KERNEL);
  dataout=kmalloc(arvid.bufflen,GFP_KERNEL);

  for(i=0; i< arvid.bufflen; i++) datainp[i]=0;

  for(bnum=0; bnum< arvid.buffers; bnum++) {
    for(i=0; i< arvid.bufflen; i++) dataout[i]=i+bnum;
    wbuf(bnum,dataout);
    rbuf(bnum,datainp);
    for(i=0; i< arvid.bufflen; i++)
      if(datainp[i] != dataout[i]) {
	err = -1;
      }
  }
  kfree(datainp);
  kfree(dataout);
  return err;
}
